#include <stdio.h>
#include <stdlib.h>
#include <conio.h>
#include <string.h>
#include <time.h>
#include <dos.h>

// -- Defines -----------------------
#define BUILD "0013 ALPHA"

#define MAX_MAPF_SIZE 255

#define MAX_MAP_XSIZE 100
#define MAX_MAP_YSIZE 100

#define MAX_VIEWS 4

#define MAX_PL 5
#define MAX_PL_LENGTH 30

#define DO_MOVE_TIME 0.05

#define MAX_CONSOLE_LINES 4

#define MAP_FILE_NAME "CTF2A.PMF"
// ----------------------------------

// Map fields
int n_mapf_size=MAX_MAPF_SIZE;
struct s_mapf {
				unsigned char m; // Map field properties
				unsigned char c; // Char
				unsigned char f,b; // Foreground, background color
			  } mapf[MAX_MAPF_SIZE];

// Game map
int n_map_xsize=MAX_MAP_XSIZE;
int n_map_ysize=MAX_MAP_YSIZE;
struct s_map {
			   char name[64]; // Map name
			   unsigned char m[MAX_MAP_XSIZE][MAX_MAP_YSIZE]; // Map fields properties
			   //unsigned char p[MAX_MAP_XSIZE][MAX_MAP_YSIZE];
			 } map;

// Player views
int n_views=0; // Number of active player views
struct s_views {
				 int xp,yp; // X and Y position on screen
				 int xm,ym; // X and Y position on map
				 int xs,ys; // X and Y size to display
				 unsigned char mv; // Moving (0..249 - pl num., 250 - don't move)
				 s_map tm1;  // Temporary map (for screen redraw)
			   } views[MAX_VIEWS];

// Player info
unsigned char n_pl=0; // Number of active players
unsigned char w_flag=255, wl_flag=255; // Who has got flag (0...250 - player, 255 - not yet stolen)
struct s_pl {
			  char name[127]; // Player name
			  unsigned long s; // Player score
			  unsigned char pcolor,phchar; // Player color, Player head character
			  unsigned char d; // Actual player direction (0=nothing, 1=up, 2=down, 3=left, 4=right)
			  unsigned char alen; //,blen; // Actual length, Absolute length(length to be)
			  float blen; // Absolute length(length to be)
			  int x[MAX_PL_LENGTH],y[MAX_PL_LENGTH]; // Player fields
			  unsigned char xsl,ysl,wsl; // X,Y and widtch of status location (on screen)
			} pl[MAX_PL];

// Console
unsigned char n_console_lines=MAX_CONSOLE_LINES;
unsigned char console_redraw=0;
struct s_console {
				   char l[127]; // Console text line
				   unsigned char f,b; // Foreground and background color
				 } console[MAX_CONSOLE_LINES];

unsigned char end_p=0;
unsigned long timer1=0,timer2=0; // Game timer

clock_t game_stime, game_atime, game_latime;

// -------------------------------------------------------------------------
void tcolor(int fc, int bc)
{
  textcolor(fc);
  textbackground(bc);
}
// -------------------------------------------------------------------------
void redraw_console()
{
  int i1;

  tcolor(8,0);
  gotoxy(1,(50-n_console_lines));
  for (i1=0; i1<80; i1++)
  {
	cprintf("");
  }

  for (i1=0; i1<n_console_lines; i1++)
  {
	gotoxy(1,(50-i1));

	tcolor(7,0);
	clreol();

	if (console[i1].f!=255) { tcolor(console[i1].f, console[i1].b); }
					   else { tcolor(15,0); }

	if (i1>0) { tcolor(8,0); }

	cprintf("%s", console[i1].l);
  }

  console_redraw=0;
}

void write_console(char text1[], unsigned char f)
{
  int i1;

  for (i1=(n_console_lines-1); i1>0; i1--)
  {
	strcpy(console[i1].l, console[(i1-1)].l);
	console[i1].f=console[(i1-1)].f;
	console[i1].b=console[(i1-1)].b;
  }

  strcpy(console[0].l, text1);
  console[0].f=f;
  console[0].b=0;

  console_redraw=1;
}

void init_console()
{
  int i1;

  for (i1=0; i1<MAX_CONSOLE_LINES; i1++)
  {
	strcpy(console[i1].l,"");
	console[i1].f=255;
	console[i1].b=0;
  }

  console_redraw=1;
}
// -------------------------------------------------------------------------
void redraw_basic_screen()
// Draw text and gfx which is not changed during game on the screen
{
  // Display some copyright information
  tcolor(7,0);
  clrscr();
  /*
  gotoxy (1,48);
  tcolor (15,0); cprintf("(c)2001 by JC & JK & JM.\r\n");
  tcolor ( 7,0); cprintf("Build \"%s\". Not for public use ;-).\r\n", BUILD);
				 cprintf("For more informations visit http://www.mobilnews.cz/worms/");
  */
}

int xy_pl_exist(int x1, int y1, int wret)
// Check if there is a player on x1,y1
// 0 - 250 ... player number
//     254 ... last field of player (not showed)
//     255 ... there is no player

// Wret
// 0 ... player num
// 1 ... field num
{
  int i1,i2;

  for (i1=0; i1<n_pl; i1++) // Players 1...n
  {
	for (i2=0; i2<=pl[i1].alen; i2++) // 0...player length
	{
	  if ( (pl[i1].x[i2]==x1) && (pl[i1].y[i2]==y1) )
	  {
		switch (wret)
		{
		  case 0 : if (i2==pl[i1].alen) { return 254; }
								   else { return i1; }
				   break;
		  case 1 : return i2;
				   break;
		}
	  }
	}
  }

  return 255;
}

unsigned char mk_random_bkgchar()
{
  if (random(100)<60) { return (32+random(220)); }
				 else { return (32); }
}

void show_map(int m_num)
// Redraw map on the screen
{
  int ix,iy,i1,i2,i3;

  // Show map
  for (iy=views[m_num].ym; iy<(views[m_num].ym+views[m_num].ys); iy++)
  for (ix=views[m_num].xm; ix<(views[m_num].xm+views[m_num].xs); ix++)
  {
	// For faster redrawing remarked and moved
	// gotoxy ( (views[m_num].xp+ix-views[m_num].xm),(views[m_num].yp+iy-views[m_num].ym) );

	if ( (ix<0) || (iy<0) || !(ix<n_map_xsize) || !(iy<n_map_ysize) )
	//   && (views[m_num].tm1.m[(ix-views[m_num].xm)][(iy-views[m_num].ym)]!=254) )
	{
	  if (views[m_num].tm1.m[(ix-views[m_num].xm)][(iy-views[m_num].ym)]!=254)
	  {
		// Scrolling line between background and map borders
		if (mapf[ views[m_num].tm1.m[(ix-views[m_num].xm)][(iy-views[m_num].ym)] ].c!='@')
		{
		  gotoxy ( (views[m_num].xp+ix-views[m_num].xm),(views[m_num].yp+iy-views[m_num].ym) );
		  tcolor ( 8, 0); cprintf("%c",mk_random_bkgchar()); // Out of map range
		  views[m_num].tm1.m[(ix-views[m_num].xm)][(iy-views[m_num].ym)]=254;
		}
	  }
	}
	else
	{
	  i3=xy_pl_exist(ix,iy,0);
	  // This is for less redrawing - draw only changed things on the screen
	  //if ((views[m_num].tm1.m[(ix-views[m_num].xm)][(iy-views[m_num].ym)]) != (map.m[ix][iy]) && (i3>250))
	  if ( ( (views[m_num].tm1.m[(ix-views[m_num].xm)][(iy-views[m_num].ym)]) != (map.m[ix][iy])
		  || ((w_flag!=wl_flag) && (mapf[ map.m[ix][iy] ].m==4)) // Flag - capture place
		  || ((w_flag!=wl_flag) && (mapf[ map.m[ix][iy] ].m==5)) // Flag - destination
		   )
		&& (i3>250) )
	  {
		gotoxy ( (views[m_num].xp+ix-views[m_num].xm),(views[m_num].yp+iy-views[m_num].ym) );

		// Flag is not yet redrawed immediately

		// Map properties
		// 0 ... free field
		// 1 ... wall
		// 2 ... free field (not yet digged)
		// 3 ... food

		tcolor(mapf[ map.m[ix][iy] ].f, mapf[ map.m[ix][iy] ].b);
		if ( (mapf[ map.m[ix][iy] ].c!='@')
		  && (mapf[ map.m[ix][iy] ].m!=4)
		  && (mapf[ map.m[ix][iy] ].m!=5) )
		{
		  cprintf("%c", mapf[ map.m[ix][iy] ].c);
		}
		else
		{
		  if (mapf[ map.m[ix][iy] ].c=='@')
		  {
			// Scrolling line between background and map borders
			if (views[m_num].tm1.m[(ix-views[m_num].xm)][(iy-views[m_num].ym)]!=254)
			{
			  cprintf("%c",mk_random_bkgchar());
			}
		  }

		  if (mapf[ map.m[ix][iy] ].m==4)
		  {
			if (w_flag==255) cprintf("%c",mapf[ map.m[ix][iy] ].c);
						else cprintf(" ");
		  }

		  if (mapf[ map.m[ix][iy] ].m==5)
		  {
			if (w_flag!=255) cprintf("%c",mapf[ map.m[ix][iy] ].c);
						else cprintf(" ");
		  }
		}

		// Write new map properties to the temporary map screen
		views[m_num].tm1.m[(ix-views[m_num].xm)][(iy-views[m_num].ym)]=map.m[ix][iy];
	  }

	  if ( (i3!=255) && (i3!=254) )
	  {
		views[m_num].tm1.m[(ix-views[m_num].xm)][(iy-views[m_num].ym)]=255;
	  }
	}
  }

  // Show players
  for (i1=0; i1<n_pl; i1++) // Players 1...n
  {
	for (i2=0; i2<pl[i1].alen; i2++) // 0...player length
	{
	  ix=pl[i1].x[i2]-views[m_num].xm;
	  iy=pl[i1].y[i2]-views[m_num].ym;

	  // If player is in view, show it
	  if ( (ix>=0) && (iy>=0) && (ix<views[m_num].xs) && (iy<views[m_num].ys) )
	  {
		ix+=views[m_num].xp;
		iy+=views[m_num].yp;
		gotoxy (ix,iy);
		tcolor(pl[i1].pcolor, 0);
		switch (i2)
		{
		   case 0 : if (w_flag==i1) { tcolor(15, pl[i1].pcolor); cprintf("F"); }
							   else { cprintf("%c",pl[i1].phchar); }
					break;
		  default : cprintf(""); break;
		}
	  }
	  else // If player is not in view, show arrow
	  {
		// To be completed ;-)
	  }
	}
  }

  gotoxy (1,1);
}

void show_pl_stat(int pln)
// Show one line players statistics
{
  int i1,i2;

  gotoxy (pl[pln].xsl,pl[pln].ysl);
  //tcolor (pl[pln].pcolor,0); cprintf("");
  tcolor (15,pl[pln].pcolor); cprintf ("  [ %s ] S=%u (", pl[pln].name, pl[pln].s);
  if (w_flag==pln) cprintf("F");
			  else cprintf("-");
  cprintf(") L=%i  ", pl[pln].alen); //, pl[pln].blen);
  i2=pl[pln].wsl-(wherex()-pl[pln].xsl);//-4;
  for (i1=0; i1<i2; i1++)
  {
	cprintf(" ");
  }
  //tcolor (pl[pln].pcolor,0); cprintf("۲ ");
}

void change_numofpl(int numofpl)
// This function will change number of active players in game
{
  int i1,i2,i3;

  switch (numofpl)
  {
	// 1 player
	case 1 : n_views=1;

			 views[0].xp=1;   views[0].yp=3;
			 views[0].xm=0;   views[0].ym=0;
			 views[0].xs=80;  views[0].ys=43;
			 views[0].mv=0;

			 n_pl=1;
			 pl[0].xsl=1; pl[0].ysl=2; pl[0].wsl=80;

			 write_console("Switched to one player mode.", 255);
			 break;

	// 2 players
	case 2 : n_views=2;

			 views[0].xp=1;   views[0].yp=3;
			 views[0].xm=0;   views[0].ym=0;
			 views[0].xs=39;  views[0].ys=43;
			 views[0].mv=0;

			 views[1].xp=42;  views[1].yp=3;
			 views[1].xm=0;   views[1].ym=0;
			 views[1].xs=39;  views[1].ys=43;
			 views[1].mv=1;

			 n_pl=2;
			 pl[0].xsl=1; pl[0].ysl=2; pl[0].wsl=39;
			 pl[1].xsl=42; pl[1].ysl=2; pl[1].wsl=39;

			 write_console("Switched to two players mode.", 255);
			 break;
	// 3 players
	case 3 : n_views=3;

			 views[0].xp=1;   views[0].yp=3;
			 views[0].xm=0;   views[0].ym=0;
			 views[0].xs=43;  views[0].ys=20;
			 views[0].mv=0;

			 views[1].xp=1;   views[1].yp=26;
			 views[1].xm=0;   views[1].ym=0;
			 views[1].xs=43;  views[1].ys=20;
			 views[1].mv=1;

			 views[2].xp=46;  views[2].yp=3;
			 views[2].xm=0;   views[2].ym=0;
			 views[2].xs=35;  views[2].ys=43;
			 views[2].mv=2;

			 n_pl=3;
			 pl[0].xsl=1;  pl[0].ysl=2; pl[0].wsl=43;
			 pl[1].xsl=1;  pl[1].ysl=25; pl[1].wsl=43;
			 pl[2].xsl=46; pl[2].ysl=2; pl[2].wsl=35;

			 write_console("Switched to three players mode.", 255);
			 break;
	// 4 players
	case 4 : n_views=4;

			 views[0].xp=1;   views[0].yp=3;
			 views[0].xm=0;   views[0].ym=0;
			 views[0].xs=39;  views[0].ys=20;
			 views[0].mv=0;

			 views[1].xp=1;   views[1].yp=26;
			 views[1].xm=0;   views[1].ym=0;
			 views[1].xs=39;  views[1].ys=20;
			 views[1].mv=1;

			 views[2].xp=42;  views[2].yp=3;
			 views[2].xm=0;   views[2].ym=0;
			 views[2].xs=39;  views[2].ys=20;
			 views[2].mv=2;

			 views[3].xp=42;  views[3].yp=26;
			 views[3].xm=0;   views[3].ym=0;
			 views[3].xs=39;  views[3].ys=20;
			 views[3].mv=3;

			 n_pl=4;
			 pl[0].xsl=1;  pl[0].ysl=2; pl[0].wsl=39;
			 pl[1].xsl=1;  pl[1].ysl=25; pl[1].wsl=39;
			 pl[2].xsl=42; pl[2].ysl=2; pl[2].wsl=39;
			 pl[3].xsl=42; pl[3].ysl=25; pl[3].wsl=39;

			 write_console("Switched to four players mode.", 255);
			 break;
  }

  for (i3=0; i3<n_pl; i3++)
  {
	for (i2=0; i2<MAX_MAP_YSIZE; i2++)
	for (i1=0; i1<MAX_MAP_XSIZE; i1++)
	{
	  views[i3].tm1.m[i1][i2]=255;
	}
  }

  console_redraw=1;
  redraw_basic_screen();
}

/*
void init_map()
// Initialise testing map
// This function will be replaced by function load_map() in future
{
  int ix,iy;

  // Map name
  strcpy(map.name,"Map1");

  // Initialise map fields
  for (iy=0; iy<n_map_ysize; iy++)
  for (ix=0; ix<n_map_xsize; ix++)
  {
	map.m[ix][iy]=0; // Clear all map fields

	// Create map borders
	if ( (ix==0) || (iy==0) || (ix==(n_map_xsize-1)) || (iy==(n_map_ysize-1)) )
	  {  map.m[ix][iy]=1;  }
  }

  // Digging fields
  for (iy=1; iy<((n_map_ysize / 2)+1); iy++)
  for (ix=(n_map_xsize-(n_map_xsize / 2)); ix<(n_map_xsize-1); ix++)
  {
	map.m[ix][iy]=2;
  }

  // Eating fields
  for (iy=((n_map_ysize / 2)+2); iy<(n_map_ysize-1); iy++)
  for (ix=((n_map_xsize / 2)+2); ix<(n_map_xsize-1); ix++)
  {
	if (random(100)<5)
	  {  map.m[ix][iy]=3;  }
  }

  // Cross in center
  for (ix=(n_map_xsize / 5); ix<((n_map_xsize / 5)*2); ix++)
  {
	map.m[ix][(n_map_ysize / 2)]=1;
	map.m[(ix+((n_map_xsize / 5)*2))][(n_map_ysize / 2)]=1;
  }

  for (iy=(n_map_ysize / 5); iy<((n_map_ysize / 5)*2); iy++)
  {
	map.m[(n_map_xsize / 2)][iy]=1;
	map.m[(n_map_xsize / 2)][(iy+((n_map_ysize / 5)*2))]=1;
  }
}
*/

void load_map(char map_file[])
{
  int i1,i2,ix,iy;
  unsigned char chx;
  FILE *in;

  printf("\n");

  printf("Map file name: %s\n", map_file);

  if ((in = fopen(map_file, "rt")) == NULL)
  {
	 printf("Cannot open %s for reading.\n", map_file);
  }
  else
  {
	// Header
	printf("Map ID: ");
	for (i1=0; i1<5; i1++)
	{
	  chx=fgetc(in);
	  printf("%c", chx);
	}
	printf("\n");

	// Map name
	strcpy(map.name, "");
	for (i1=0; i1<32; i1++)
	{
	  chx=fgetc(in);
	  map.name[i1]=chx;
	}
	map.name[32]=0;
	printf("Map name: %s\n", map.name);

	// Mapf size
	n_mapf_size=fgetc(in); // MAPF_SIZE

	printf("mapf_size: %i\n", n_mapf_size);

	// Map size
	n_map_xsize=fgetc(in); // X_SIZE
	n_map_ysize=fgetc(in); // Y_SIZE

	printf("x_size / max_x_size: %i / %i\n", n_map_xsize, MAX_MAP_XSIZE);
	printf("y_size / max_y_size: %i / %i\n", n_map_ysize, MAX_MAP_YSIZE);

	// Players
	i2=fgetc(in); // Number of players
	for (i1=0; i1<i2; i1++)
	{
	  pl[i1].x[0]=fgetc(in);
	  pl[i1].y[0]=fgetc(in);
	}
	printf("Players: %i\n", i2);

	for (i1=0; i1<n_mapf_size; i1++)
	{
	  mapf[i1].m=fgetc(in);
	  mapf[i1].c=fgetc(in);
	  mapf[i1].f=fgetc(in);
	  mapf[i1].b=fgetc(in);
	}

	if ( (n_map_xsize<=MAX_MAP_XSIZE) && (n_map_ysize<=MAX_MAP_YSIZE) )
	{
	  for (iy=0; iy<n_map_ysize; iy++)
	  for (ix=0; ix<n_map_xsize; ix++)
	  {
		map.m[ix][iy]=fgetc(in);
	  }

	  printf("OK, map successfully loaded.\n");
	}
	else
	{
	  printf("Map size is out of range.\n");
	}

	fclose(in);
  }

  printf("\n\nPress any key...");
  getch();
}

/*
void init_mapf()
{
  // m=0 ... normal playing field
  // m=1 ... wall
  // m=2 ... tuneler playing filed
  // m=3 ... eat

  mapf[0].m= 0; mapf[0].f= 1; mapf[0].b= 1; mapf[0].c=' ';
  mapf[1].m= 1; mapf[1].f=14; mapf[1].b=14; mapf[1].c='';
  mapf[2].m= 2; mapf[2].f= 9; mapf[2].b= 1; mapf[2].c='';
  mapf[3].m= 3; mapf[3].f=10; mapf[3].b= 1; mapf[3].c='O';
}
*/

void init_pl()
// Initialise testing players
{
  n_pl=2;

  // Player1
  strcpy(pl[0].name,"Player1");
  pl[0].s=0;
  pl[0].pcolor=11;
  pl[0].phchar='';
  pl[0].d=0;
  pl[0].alen=1; pl[0].blen=5.0;
  //pl[0].x[0]=5; pl[0].y[0]=5;

  // Player2
  strcpy(pl[1].name,"Player2");
  pl[1].s=0;
  pl[1].pcolor=10;
  pl[1].phchar='';
  pl[1].d=0;
  pl[1].alen=1; pl[1].blen=5.0;
  //pl[1].x[0]=n_map_xsize-7; pl[1].y[0]=n_map_ysize-7;

  // Player3
  strcpy(pl[2].name,"Player3");
  pl[2].s=0;
  pl[2].pcolor=12;
  pl[2].phchar='';
  pl[2].d=0;
  pl[2].alen=1; pl[2].blen=5.0;
  //pl[2].x[0]=n_map_xsize-7; pl[2].y[0]=5;

  // Player4
  strcpy(pl[3].name,"Player4");
  pl[3].s=0;
  pl[3].pcolor=13;
  pl[3].phchar='';
  pl[3].d=0;
  pl[3].alen=1; pl[3].blen=5.0;
  //pl[2].x[0]=n_map_xsize-7; pl[2].y[0]=5;
}

void init_views()
// Initialise views
// This function is now handled by function change_numofpl()
{
  // Init some testing map views
  /*
  n_views=2;
  views[0].xp=1;   views[0].yp=3;
  views[0].xm=0;   views[0].ym=0;
  views[0].xs=38;  views[0].ys=10;
  views[0].mv=0;

  views[1].xp=40;  views[1].yp=3;
  views[1].xm=0;   views[1].ym=0;
  views[1].xs=38;  views[1].ys=10;
  views[1].mv=1;
  */
  change_numofpl(1);
}

// -------------------------------------------------------------------------

void get_keyb()
// This function handle everything about keyboard
{
  char ch;

  if (kbhit())
  {
	ch=getch();
	switch (ch)
	{
	  case 27 : end_p=1; break; // ESC

	  // Player 1 changes and debug
	  case 'x' : pl[0].d=0; break; // Pause player pl[0]
	  case 'c' : pl[0].blen++; break; // Add length to pl[0]
	  case 'z' : pl[0].blen--; break; // Remove length from pl[0]
	  case 'r' : game_stime=clock();
				 game_atime=clock();
				 game_latime=clock()-(DO_MOVE_TIME / CLK_TCK)-1;
				 timer1=0;
				 timer2=0;
				 break;
	  case 't' : write_console("test1234", 255);
				 break;

	  // Player 2 controls
	  case 'w' : if (pl[1].d!=2) pl[1].d=1; break; // Up arrow
	  case 's' : if (pl[1].d!=1) pl[1].d=2; break; // Down arrow
	  case 'a' : if (pl[1].d!=4) pl[1].d=3; break; // Left arrow
	  case 'd' : if (pl[1].d!=3) pl[1].d=4; break; // Right arrow

	  // Player 3 controls
	  case 'i' : if (pl[2].d!=2) pl[2].d=1; break; // Up arrow
	  case 'k' : if (pl[2].d!=1) pl[2].d=2; break; // Down arrow
	  case 'j' : if (pl[2].d!=4) pl[2].d=3; break; // Left arrow
	  case 'l' : if (pl[2].d!=3) pl[2].d=4; break; // Right arrow

	   case 0 : ch=getch();
				switch (ch)
				{
				  case 72 : if (pl[0].d!=2) pl[0].d=1; break; // Up arrow
				  case 80 : if (pl[0].d!=1) pl[0].d=2; break; // Down arrow
				  case 75 : if (pl[0].d!=4) pl[0].d=3; break; // Left arrow
				  case 77 : if (pl[0].d!=3) pl[0].d=4; break; // Right arrow

				  case 82 : change_numofpl(2); break; // INS
				  case 83 : change_numofpl(1); break; // DEL
				  case 71 : change_numofpl(3); break; // HOME
				  case 79 : change_numofpl(4); break; // END
				}
				ch=0;
				break;
	}
  }
}

void redraw_screen()
{
  int i1;

  for (i1=0; i1<n_views; i1++)
  {
	show_map(i1);
  }

  for (i1=0; i1<n_pl; i1++)
  {
	show_pl_stat(i1);
  }

  wl_flag=w_flag;
}

void do_game()
{
  int i1,i2,i3,i4,i5,c_move,new_x,new_y;
  char s1[127];

  for (i1=0; i1<n_pl; i1++) // Players 1...n
  {
	new_x=pl[i1].x[0];
	new_y=pl[i1].y[0];

	c_move=1;
	switch (pl[i1].d)
	{
	   // Up
	   case 1 : if ( !(mapf[ map.m[ pl[i1].x[0] ][ (pl[i1].y[0]-1) ]].m!=1)
				  || (pl[i1].y[0]<=0) ) // || (xy_pl_exist(pl[i1].x[0],(pl[i1].y[0]-1),0)<250) )
				  { c_move=0; }
				new_x=pl[i1].x[0];
				new_y=pl[i1].y[0]-1;
				break;

	   // Down
	   case 2 : if ( !(mapf[ map.m[ pl[i1].x[0] ][ (pl[i1].y[0]+1) ]].m!=1)
				  || (pl[i1].y[0]>=(n_map_ysize-1)) ) //|| (xy_pl_exist(pl[i1].x[0],(pl[i1].y[0]+1),0)<250) )
				  { c_move=0; }
				new_x=pl[i1].x[0];
				new_y=pl[i1].y[0]+1;
				break;

	   // Left
	   case 3 : if ( !(mapf[ map.m[ (pl[i1].x[0]-1) ][ pl[i1].y[0] ]].m!=1)
				  || (pl[i1].x[0]<=0) ) // || (xy_pl_exist((pl[i1].x[0]-1),pl[i1].y[0],0)<250) )
				  { c_move=0; }
				new_x=pl[i1].x[0]-1;
				new_y=pl[i1].y[0];
				break;

	   // Right
	   case 4 : if ( !(mapf[ map.m[ (pl[i1].x[0]+1) ][ pl[i1].y[0] ]].m!=1)
				  || (pl[i1].x[0]>=(n_map_xsize-1)) ) // || (xy_pl_exist((pl[i1].x[0]+1),pl[i1].y[0],0)<250) )
				  { c_move=0; }
				new_x=pl[i1].x[0]+1;
				new_y=pl[i1].y[0];
				break;

	  // Unknown direction ... it will be better if we will not move ;-)
	  default : c_move=0; break;
	}

	// Add or remove length
	if ( (pl[i1].alen<int(pl[i1].blen)) && (c_move==1) ) pl[i1].alen++;
	if (pl[i1].alen>int(pl[i1].blen)) pl[i1].alen--;

	if (pl[i1].alen<1) { pl[i1].alen=1; pl[i1].blen=1.0; }
	if (pl[i1].alen>(MAX_PL_LENGTH-1)) { pl[i1].alen=(MAX_PL_LENGTH-1); pl[i1].blen=(float(MAX_PL_LENGTH)-1.0); }

	// Change other worm prop.
	i2=xy_pl_exist(new_x,new_y,0); // Player number
	i3=xy_pl_exist(new_x,new_y,1); // Player filed num
	if ( (i2<250) && (i2!=i1) )
	{
	  i5=(pl[i2].alen-i3);
	  if ( !((pl[i1].alen+i5)<MAX_PL_LENGTH) )
		{ i5=MAX_PL_LENGTH-pl[i1].alen-1; }
	  for (i4=0; i4<i5; i4++)
	  {
		pl[i1].x[ (pl[i1].alen+i4) ]=pl[i2].x[(i3+i4)];
		pl[i1].y[ (pl[i1].alen+i4) ]=pl[i2].y[(i3+i4)];
	  }
	  pl[i1].alen=(pl[i1].alen+i5);
	  pl[i1].blen=(pl[i1].alen+i5);

	  pl[i2].alen=i3;
	  pl[i2].blen=i3;

	  if (w_flag==i1)
	  {
		w_flag=i2;

		// Console message
		strcpy(s1, pl[i2].name);
		strcat(s1, " received the flag from ");
		strcat(s1, pl[i1].name);
		strcat(s1, ".");
		write_console(s1, pl[i2].pcolor);
	  }
	}

	// If move is possible, worm will move to the new position
	if (c_move==1)
	{
	  // Finally, we move the worm
	  for (i2=(pl[i1].alen-1); i2>0; i2--)
	  {
		pl[i1].x[i2]=pl[i1].x[(i2-1)];
		pl[i1].y[i2]=pl[i1].y[(i2-1)];
	  }
	  pl[i1].x[0]=new_x;
	  pl[i1].y[0]=new_y;
	}
	else
	{
	  // If player is not moving - it will lost the flag
	  if (w_flag==i1)
	  {
		w_flag=255;

		// Console message
		strcpy(s1, pl[i1].name);
		strcat(s1, " lost the flag.");
		write_console(s1, pl[i1].pcolor);
	  }
	}

	// Changing length with properties of map field
	switch (mapf[ map.m[ (pl[i1].x[0]) ][ pl[i1].y[0] ] ].m)
	{
	  //case 0 : pl[i1].blen=pl[i1].blen-0.01; break;
	  case 2 : pl[i1].blen=pl[i1].blen+0.1; break;
	  case 3 : pl[i1].blen++; break;
	}

	// Player captured the flag
	if ( (mapf[ map.m[ (pl[i1].x[0]) ][ pl[i1].y[0] ] ].m==4) && (w_flag==255) )
	{
	  w_flag=i1;

	  // Console message
	  strcpy(s1, pl[i1].name);
	  strcat(s1, " captured the flag.");
	  write_console(s1, pl[i1].pcolor);
	}

	// Flag arrived to destination
	if ( (mapf[ map.m[ (pl[i1].x[0]) ][ pl[i1].y[0] ] ].m==5) && (w_flag==i1) )
	{
	  w_flag=255;
	  pl[i1].s=pl[i1].s+100;

	  // Console message
	  strcpy(s1, pl[i1].name);
	  strcat(s1, " scored (moved flag to the destination).");
	  write_console(s1, pl[i1].pcolor);
	}

	if ( (mapf[ map.m[ (pl[i1].x[0]) ][ pl[i1].y[0] ] ].m!=0)
	  && (mapf[ map.m[ (pl[i1].x[0]) ][ pl[i1].y[0] ] ].m!=4)
	  && (mapf[ map.m[ (pl[i1].x[0]) ][ pl[i1].y[0] ] ].m!=5) )
	{
	  map.m[ (pl[i1].x[0]) ][ pl[i1].y[0] ]=0;
	}
  }

  // Move views
  for (i1=0; i1<n_views; i1++)
  {
	if (views[i1].mv<250)
	{
	  // Move view with player
	  views[i1].xm=pl[ views[i1].mv ].x[0]-(views[i1].xs / 2);
	  views[i1].ym=pl[ views[i1].mv ].y[0]-(views[i1].ys / 2);

	  // If we are on the end, don't move view
	  /*
	  if (views[i1].xm<0) views[i1].xm=0;
	  if (views[i1].ym<0) views[i1].ym=0;
	  if ((views[i1].xm+views[i1].xs)>(n_map_xsize-1)) views[i1].xm=n_map_xsize-views[i1].xs;
	  if ((views[i1].ym+views[i1].ys)>(n_map_ysize-1)) views[i1].ym=n_map_ysize-views[i1].ys;
	  */
	}
  }
}

void show_version_info()
{
  tcolor(7,0);
  clrscr();

  tcolor(15,0);
  cprintf("Basic control info:\r\n");
  tcolor(11,0); cprintf("Player1 - arror keys + Z(decrease length) + X(stop worm) + C(increase length)\r\n");
  tcolor(10,0); cprintf("Player2 - W(up) + S(down) + A(left) + D(right)\r\n");
  tcolor(12,0); cprintf("Player3 - I(up) + K(down) + J(left) + L(right)\r\n");
  tcolor(13,0); cprintf("Player4 - controls not yet implemented ;-)\r\n\n");

  tcolor(15,0);
  cprintf("Other control info:\r\n");
  tcolor(7,0);
  cprintf("DELETE - switch to one player mode (default)\r\n");
  cprintf("INSERT - switch to two players mode\r\n");
  cprintf("HOME   - switch to three players mode\r\n");
  cprintf("END    - switch to four players mode\r\n");
  cprintf("R      - reset timer\r\n\n");

  tcolor(7,0);
}

void main()
{
  char s1[127];

  textmode(C4350);
  randomize();

  show_version_info();

  //init_map();
  load_map(MAP_FILE_NAME);
  init_pl();
  init_views();
  init_console();

  // -- Console - welcome info --------------------------------
  write_console("(c)2001 by JC & JK & JM.", 255);

  strcpy(s1, "Build \"");
  strcat(s1, BUILD);
  strcat(s1, "\". Not for public user ;-).");
  write_console(s1, 255);

  write_console("For more informations visit http://www.mobilnews.cz/worms/", 255);
  write_console("", 255);
  // ----------------------------------------------------------

  redraw_basic_screen();

  // Initialise game timer (start)
  game_stime=clock();

  // Initialise game timer (actual time and last actual time)
  game_atime=clock();
  game_latime=clock()-(DO_MOVE_TIME / CLK_TCK)-1;

  while (end_p==0)
  {
	get_keyb();

	game_atime=clock();
	if ( (game_atime / CLK_TCK) > (((game_latime) / CLK_TCK)+DO_MOVE_TIME) )
	{
	  game_latime=clock();
	  timer2++;

	  do_game();
	  //redraw_screen();

	  // Division by zero
	  if (((game_atime-game_stime) / CLK_TCK) != 0)
	  {
		gotoxy (1,1);
		printf("Time: %0.5f  ", ((game_atime-game_stime) / CLK_TCK));
		printf("FPS: %0.2f  Frame:%u  ", (float(timer1) / ((game_atime-game_stime) / CLK_TCK)), timer1);
		printf("MPS: %0.2f (%0.2f%%)  Moves:%u  ", (float(timer2) / ((game_atime-game_stime) / CLK_TCK)), (100.0 / (1.0 / DO_MOVE_TIME))*(float(timer2) / ((game_atime-game_stime) / CLK_TCK)) , timer2);
	  }
	}

	redraw_screen();
	if (console_redraw==1) { redraw_console(); }

	timer1++;
  }

  tcolor(7,0);
  gotoxy(1,50);
}
